###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Relative package containing all things handling GUI elements."""

__all__ = (
    'AccountContextMenu',
    'app',
    'AppWindow',
    'ComboBox',
    'ConversationView',
    'ExceptionLogger',
    'ExceptionReporter',
    'ExternalTextBrowser',
    'GetterApp',
    'HelpContextMenu',
    'HistoryComboBox',
    'LicenseViewer',
    'PasteLineEdit',
    'ReadmeViewer',
    'SettingsWindow',
    'SignInDialog',
    'Theme',
    'ToolsContextMenu',
    'tr',
    'TranslatableComboBox',
)

from .aliases import app
from .aliases import tr
from .app import GetterApp
from .app import Theme
from .menus import AccountContextMenu
from .menus import HelpContextMenu
from .menus import ToolsContextMenu
from .widgets import ComboBox
from .widgets import ConversationView
from .widgets import ExceptionLogger
from .widgets import ExternalTextBrowser
from .widgets import HistoryComboBox
from .widgets import PasteLineEdit
from .widgets import TranslatableComboBox
from .windows import AppWindow
from .windows import ChangelogViewer
from .windows import ExceptionReporter
from .windows import LicenseViewer
from .windows import ReadmeViewer
from .windows import SettingsWindow
from .windows import SignInDialog



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Defines the Client structures."""
from __future__ import annotations

__all__ = (
    'Action',
    'Conversation',
    'Message',
    'User',
)

from dataclasses import dataclass
from dataclasses import field
from typing import Any
from uuid import UUID
from uuid import uuid4


@dataclass
class Message:
    """ChatGPT message sent or received."""

    uuid: UUID = field(default_factory=uuid4)
    text: str | None = field(default=None)
    role: str = field(default='user')

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Message:
        """Load data from a JSON representation."""
        return cls(
            uuid=UUID(data['id']),
            text='\n\n'.join(data['content']['parts']),
            role=data['role']
        )

    def to_json(self) -> dict[str, Any]:
        """Dump data into a JSON representation.

        The ``content`` key is omitted if the ``text`` attribute is ``None``.
        """
        data: dict[str, Any] = {
            'id': str(self.uuid),
            'role': self.role
        }

        if self.text is not None:
            data['content'] = {
                'content_type': 'text',
                'parts': [self.text]
            }

        return data


@dataclass
class Conversation:
    """ChatGPT conversation.

    Includes a list of all messages sent and received in the conversation.
    """

    uuid: UUID | None = field(default=None)
    messages: list[Message] = field(default_factory=list)

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> Conversation:
        """Load data from a JSON representation."""
        return cls(
            uuid=UUID(data['id']),
            messages=[Message.from_json(message) for message in data['messages']],
        )

    def to_json(self) -> dict[str, Any]:
        """Dump data into a JSON representation."""
        return {
            'id': str(self.uuid),
            'messages': [message.to_json() for message in self.messages],
        }


@dataclass
class Action:
    """Action sent to ChatGPT.

    If no conversation is provided, ChatGPT creates a new conversation for you in its response.
    """

    model: str
    conversation: Conversation = field(default_factory=Conversation)
    messages: list[Message] = field(default_factory=list)
    parent: Message = field(default_factory=Message)
    type: str = field(default='next')

    def to_json(self) -> dict[str, Any]:
        """Dump data into a JSON representation.

        The ``conversation_id`` key is omitted if the ``conversation`` attribute is ``None``.
        """
        data = {
            'action': self.type,
            'messages': [message.to_json() for message in self.messages],
            'model': self.model,
            'parent_message_id': str(self.parent.uuid)
        }

        if self.conversation.uuid is not None:
            data['conversation_id'] = str(self.conversation.uuid)

        return data


@dataclass
class User:
    """User that is signed-in to ChatGPT."""

    id: str
    name: str
    email: str
    image: str
    picture: str
    groups: list[str] = field(default_factory=list)
    features: list[str] = field(default_factory=list)

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> User:
        """Load data from a JSON representation."""
        return cls(**data)

    def to_json(self) -> dict[str, Any]:
        """Dump data into a JSON representation."""
        return {
            'id': self.id,
            'name': self.name,
            'email': self.email,
            'image': self.image,
            'picture': self.picture,
            'groups': self.groups,
            'features': self.features
        }



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Defines the OpenAI authenticator."""
from __future__ import annotations

__all__ = (
    'Authenticator',
)

import re
import base64
from typing import Final
from urllib.parse import quote

from bs4 import BeautifulSoup
from PySide6.QtCore import *
from PySide6.QtGui import *
from tls_client import Session

from ...constants import *
from .structures import User


_STATE_PATTERN: Final[re.Pattern] = re.compile(r'state=(.*)')


class Authenticator(QObject):
    """Object which handles authenticating users to OpenAI.

    Based on https://github.com/rawandahmad698/PyChatGPT
    """

    authenticationFailed = Signal(str, Exception)
    """Emits the username and exception when Authenticator.authenticate() raises an Exception"""

    authenticationSuccessful = Signal(str, User)
    """Emits the session_token on success of Authenticator.authenticate()."""

    captchaEncountered = Signal(QPixmap)
    """Emits a captcha that needs to be solved."""

    solveCaptcha = Signal(str)
    """Emit this signal with the solved captcha from captchaEncountered."""

    def __init__(self, parent: QObject | None = None, username: str | None = None, password: str | None = None):
        """Create a new :py:class:`Authenticator`.

        username and password attributes must be filled out before calling ``authenticate()``.

        :param parent: Parent of QObject instance.
        :param username: Username to login to (email address).
        :param password: Password associated with username.
        """
        super().__init__(parent)

        self.username: str | None = username
        self.password: str | None = password
        self.session = Session(client_identifier='chrome_105')

    def _authenticate(self) -> None:
        """Run all steps.

        :raises ValueError: If Email or password is not provided for Authenticator.
            If csrf token was not provided by endpoint call.
            If auth token was not provided at end of function.

        :raises RuntimeError: If we couldn't get the login page.
            If we couldn't start auth0 login. Could be due to excessive login attempts.
            If there was an unsuccessful request to url provided by signin page.
            If we couldn't access the login state.
            If the email address is not a valid user.
            If password or captcha was wrong.
        """
        # -----------------------------------------------------------------------------------------
        # Step 0:
        # Check that all attributes are valid
        # -----------------------------------------------------------------------------------------
        if not self.username or not self.password:
            raise ValueError('Email or password is not provided for Authenticator.')

        username_encoded: str = quote(self.username)
        password_encoded: str = quote(self.password)

        # -----------------------------------------------------------------------------------------
        # Step 1:
        # Make a request to https://chat.openai.com/auth/login
        # This will block us (403 Request Blocked) early if detected as bot.
        # -----------------------------------------------------------------------------------------
        if self.session.get(url='https://chat.openai.com/auth/login', headers={
            'Host': 'ask.openai.com',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'User-Agent': CG_USER_AGENT,
            'Accept-Language': 'en-GB,en-US;q=0.9,en;q=0.8',
            'Accept-Encoding': 'gzip, deflate, br',
            'Connection': 'keep-alive',
        }).status_code != 200:
            raise RuntimeError('Could not get login page.')

        # -----------------------------------------------------------------------------------------
        # Step 2:
        # Get csrf token from endpoint
        # -----------------------------------------------------------------------------------------
        response = self.session.get(url='https://chat.openai.com/api/auth/csrf', headers={
            'Host': 'ask.openai.com',
            'Accept': '*/*',
            'Connection': 'keep-alive',
            'User-Agent': CG_USER_AGENT,
            'Accept-Language': 'en-GB,en-US;q=0.9,en;q=0.8',
            'Referer': 'https://chat.openai.com/auth/login',
            'Accept-Encoding': 'gzip, deflate, br',
        })

        if 'json' not in response.headers['Content-Type'] or (csrf := response.json()['csrfToken']) is None:
            raise ValueError('csrf token was not provided by endpoint call.')

        # -----------------------------------------------------------------------------------------
        # Step 3:
        # Start auth0 signin process with csrf token
        # This directs us a new url to follow
        # -----------------------------------------------------------------------------------------
        response = self.session.post(url='https://chat.openai.com/api/auth/signin/auth0?prompt=login', headers={
            'Host': 'ask.openai.com',
            'Origin': 'https://chat.openai.com',
            'Connection': 'keep-alive',
            'Accept': '*/*',
            'User-Agent': CG_USER_AGENT,
            'Referer': 'https://chat.openai.com/auth/login',
            'Content-Length': '100',
            'Accept-Language': 'en-GB,en-US;q=0.9,en;q=0.8',
            'Content-Type': 'application/x-www-form-urlencoded',
        }, data=f'callbackUrl=%2F&csrfToken={csrf}&json=true')

        if response.status_code != 200 or 'json' not in response.headers['Content-Type']:
            raise RuntimeError('Could not start auth0 login. Could be due to excessive login attempts.')

        new_url: str = response.json()['url']

        # -----------------------------------------------------------------------------------------
        # Step 4:
        # Follow new url to obtain our state key
        # -----------------------------------------------------------------------------------------
        response = self.session.get(url=new_url, headers={
            'Host': 'auth0.openai.com',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'Connection': 'keep-alive',
            'User-Agent': CG_USER_AGENT,
            'Accept-Language': 'en-US,en;q=0.9',
            'Referer': 'https://chat.openai.com/',
        })

        if response.status_code != 302:
            raise RuntimeError(f'Unsuccessful request to given url "{new_url}".')

        state: str = _STATE_PATTERN.findall(response.text)[0].split('"')[0]

        # -----------------------------------------------------------------------------------------
        # Step 5:
        # Go to our state's login page
        # This has a chance to give us a captcha to solve
        # -----------------------------------------------------------------------------------------
        response = self.session.get(f'https://auth0.openai.com/u/login/identifier?state={state}', headers={
            'Host': 'auth0.openai.com',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'Connection': 'keep-alive',
            'User-Agent': CG_USER_AGENT,
            'Accept-Language': 'en-US,en;q=0.9',
            'Referer': 'https://chat.openai.com/',
        })

        if response.status_code != 200:
            raise RuntimeError('Couldn\'t make request to our login page.')

        captcha: str | None = None
        soup: BeautifulSoup = BeautifulSoup(response.text, 'lxml')
        if soup.find('img', alt='captcha'):
            captcha_svg: str = soup.find('img', alt='captcha')['src'].split(',')[1]  # type: ignore
            decoded_svg: bytes = base64.decodebytes(captcha_svg.encode('ascii'))

            pixmap: QPixmap = QPixmap()
            pixmap.loadFromData(decoded_svg)
            captcha = self.handle_captcha(pixmap)

        # -----------------------------------------------------------------------------------------
        # Step 6:
        # Check if the username (email address) is a valid user
        # Also send the solved captcha if given
        # -----------------------------------------------------------------------------------------
        payload: str = f'state={state}&username={username_encoded}&js-available=false' \
                       f'&webauthn-available=true&is-brave=false&webauthn-platform-available=true&action=default'

        if captcha is not None:
            payload = f'state={state}&username={username_encoded}&captcha={captcha}&js-available=true' \
                      f'&webauthn-available=true&is-brave=false&webauthn-platform-available=true&action=default'

        response = self.session.post(f'https://auth0.openai.com/u/login/identifier?state={state}', headers={
            'Host': 'auth0.openai.com',
            'Origin': 'https://auth0.openai.com',
            'Connection': 'keep-alive',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'User-Agent': CG_USER_AGENT,
            'Referer': f'https://auth0.openai.com/u/login/identifier?state={state}',
            'Accept-Language': 'en-US,en;q=0.9',
            'Content-Type': 'application/x-www-form-urlencoded',
        }, data=payload)

        if response.status_code != 302:
            raise RuntimeError('Email was not a valid user.')

        # -----------------------------------------------------------------------------------------
        # Step 7:
        # Send the email and password to login to the account
        # This will give us our final state in return
        # -----------------------------------------------------------------------------------------
        payload = f'state={state}&username={username_encoded}&password={password_encoded}&action=default'
        response = self.session.post(f'https://auth0.openai.com/u/login/password?state={state}', headers={
            'Host': 'auth0.openai.com',
            'Origin': 'https://auth0.openai.com',
            'Connection': 'keep-alive',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'User-Agent': CG_USER_AGENT,
            'Referer': f'https://auth0.openai.com/u/login/password?state={state}',
            'Accept-Language': 'en-US,en;q=0.9',
            'Content-Type': 'application/x-www-form-urlencoded',
        }, data=payload)

        if response.status_code != 302:
            raise RuntimeError('Password or captcha was wrong.')

        new_state: str = _STATE_PATTERN.findall(response.text)[0].split('"')[0]

        # -----------------------------------------------------------------------------------------
        # Step 8:
        # Finish auth0 process by sending the finished state
        # This will give us our session token in return
        # -----------------------------------------------------------------------------------------
        response = self.session.get(f'https://auth0.openai.com/authorize/resume?state={new_state}', headers={
            'Host': 'auth0.openai.com',
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'Connection': 'keep-alive',
            'User-Agent': CG_USER_AGENT,
            'Accept-Language': 'en-GB,en-US;q=0.9,en;q=0.8',
            'Referer': f'https://auth0.openai.com/u/login/password?state={state}',
        }, allow_redirects=True)

        if response.status_code != 200:
            raise RuntimeError('Couldn\'t resume authorization state.')

        if not (session_token := self.session.cookies.get('__Secure-next-auth.session-token')):
            raise ValueError('While most of the process was successful, Auth0 didn\'t issue a session token, retry.')

        # -----------------------------------------------------------------------------------------
        # Test success by calling the session endpoint:
        # Emit the session token along with the session's User
        # -----------------------------------------------------------------------------------------
        response = self.session.get('https://chat.openai.com/api/auth/session', headers={
            'Host': 'chat.openai.com',
            'Accept': '*/*',
            'Connection': 'keep-alive',
            'User-Agent': CG_USER_AGENT,
            'Accept-Language': 'en-GB,en-US;q=0.9,en;q=0.8',
        })

        user: User = User.from_json(response.json()['user'])
        self.authenticationSuccessful.emit(session_token, user)

    def authenticate(self) -> None:
        """Authenticate self to OpenAI using email and password.

        Steps:
            1. Get login page
            2. Get csrf token from endpoint
            3. Post csrf token to signin page
            4. Get state from url returned by login page
            5. Solve CAPTCHA
            6. Post email to signin page
            7. Post email and password to login page to get new state
            8. Resume authorization to finish
        """
        try:
            self._authenticate()
        except Exception as e:
            self.authenticationFailed.emit(self.username, e)

    def handle_captcha(self, image: QPixmap) -> str:
        """Wait for the application implementation to finish the captcha.

        :param image: Captcha image to solve.
        """
        self.solveCaptcha.connect((holder := []).append)
        self.captchaEncountered.emit(image)

        while not holder:
            QCoreApplication.processEvents()

        return holder[0]



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Main networking package for ChatGPT-GUI."""

__all__ = (
    'Action',
    'Authenticator',
    'Client',
    'Conversation',
    'Message',
    'User',
)

from .auth import Authenticator
from .chatgpt import Client
from .structures import Action
from .structures import Conversation
from .structures import Message
from .structures import User



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Implementations for worker-thread runnables."""
from __future__ import annotations

__all__ = ()

from collections.abc import Callable

from PySide6.QtCore import *
from shiboken6 import Shiboken


class _SignalHolder(QObject):
    exceptionRaised = Signal(Exception)
    valueReturned = Signal(object)


class _Worker(QRunnable):
    _signal_holder: type[_SignalHolder] = _SignalHolder

    def __init__(self, **kwargs: Callable | Slot) -> None:
        super().__init__()
        self.signals = self._signal_holder()

        # Connect signals from keyword arguments
        for kw, val in kwargs.items():
            if hasattr(self.signals, kw) and isinstance((signal := getattr(self.signals, kw)), SignalInstance):
                signal.connect(val)
            else:
                raise TypeError(f'"{kw}" is not a valid kwarg or signal name.')

    # pylint: disable=no-self-use
    def _dummy_method(self) -> None:
        return

    def _run(self) -> None:
        raise NotImplementedError

    # pylint: disable=broad-except
    @Slot()
    def run(self) -> None:
        """Ran by the :py:class:`QThreadPool`.

        Sends non-``None`` return values through the ``valueReturned`` signal.
            - If you need to return ``None``, I suggest creating a separate object to represent it.

        Sends any uncaught :py:class:`Exception`'s through the ``exceptionRaised`` signal.

        :raises RuntimeError: If worker started before application instance is defined.
        """
        if (app := QCoreApplication.instance()) is None:
            raise RuntimeError('Worker started before application instance is defined.')

        # No idea how, but this fixes application deadlock cause by RecursiveSearch (issue #31)
        app.aboutToQuit.connect(  # pyright: ignore[reportGeneralTypeIssues]
            self._dummy_method, Qt.ConnectionType.BlockingQueuedConnection
        )

        try:
            # Emit non-None return values through the `valueReturned` signal.
            if (ret_val := self._run()) is not None:
                self.signals.valueReturned.emit(ret_val)

        except Exception as e:
            # This occurs when the application is exiting and an internal C++ object is being read after it is deleted.
            # So, return quietly and allow the process to exit with no errors.
            if not Shiboken.isValid(self.signals):
                return

            self.signals.exceptionRaised.emit(e)

        # Disconnect deadlock safeguard if successful to avoid possible leak
        app.aboutToQuit.disconnect(self._dummy_method)  # pyright: ignore[reportGeneralTypeIssues]

        # Delete if not deleted
        if Shiboken.isValid(self.signals):
            self.signals.deleteLater()



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Main application window implementation."""
from __future__ import annotations

__all__ = (
    'AppWindow',
    'size_label_for',
)

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ...constants import *
from ...events import EventBus
from ...exception_hook import ExceptionEvent
from ...models import DeferredCallable
from ...models import DistributedCallable
from ...models import Singleton
from ...utils import init_layouts
from ...utils import init_objects
from ..aliases import app
from ..menus import AccountContextMenu
from ..menus import HelpContextMenu
from ..menus import ToolsContextMenu
from ..widgets import ConversationView
from ..widgets import ExceptionLogger


def size_label_for(num: int) -> str:
    """Return the best display unit to describe the given data's size.

    Ex: Bytes, KiB, MiB, GiB, TiB
    """
    display_unit = 'Bytes'
    for size_label, size in BYTE_UNITS.items():
        if num >= (size // 2):
            display_unit = size_label
        else:
            break
    return display_unit


class AppWindow(Singleton, QMainWindow):
    """Main window for the ChatGPT-GUI application."""

    _singleton_base_type = QMainWindow
    _singleton_check_ref = False

    def __init__(self, size: QSize) -> None:
        """Create the window for the application."""
        super().__init__()

        self.resize(size)
        self.conversation_counter: int = 0

        self._init_toolbar()
        self._init_ui()

    def _init_toolbar(self) -> None:
        """Initialize toolbar widgets."""
        from .settings import SettingsWindow

        def context_menu_handler(menu_class: type[QMenu]) -> None:
            """Create a new :py:class:`QMenu` and show it at the cursor's position.

            :raises TypeError: If menu_class is not an instance of QMenu.
            """
            if not issubclass(menu_class, QMenu):
                raise TypeError(f'{menu_class} is not a subclass of {QMenu}')

            menu: QMenu = menu_class(self)
            menu.exec(self.cursor().pos())
            menu.deleteLater()

        init_objects({
            (menu_bar := QToolBar(self)): {},

            (status_bar := QToolBar(self)): {
                'movable': False,
            },

            (account := QAction(self)): {
                'icon': app().icon_store['account'],
                'menuRole': QAction.MenuRole.AboutRole,
                'triggered': DeferredCallable(context_menu_handler, AccountContextMenu)
            },

            (settings := QAction(self)): {
                'menuRole': QAction.MenuRole.PreferencesRole,
                'triggered': DistributedCallable((
                    SettingsWindow.instance().show,            # pyright: ignore[reportGeneralTypeIssues]
                    SettingsWindow.instance().activateWindow,  # pyright: ignore[reportGeneralTypeIssues]
                    SettingsWindow.instance().raise_           # pyright: ignore[reportGeneralTypeIssues]
                ))
            },

            (tools := QAction(self)): {
                'menuRole': QAction.MenuRole.ApplicationSpecificRole,
                'triggered': DeferredCallable(context_menu_handler, ToolsContextMenu)
            },

            (help := QAction(self)): {
                'menuRole': QAction.MenuRole.AboutRole,
                'triggered': DeferredCallable(context_menu_handler, HelpContextMenu)
            },

            (logger := ExceptionLogger(self)): {
                'size': {'fixed': (None, 20)},
                'clicked': DistributedCallable((
                    logger.reporter.show,
                    logger.reporter.activateWindow,
                    logger.reporter.raise_
                ))
            },
        })

        # app().client.authenticator.authenticationSuccessful.connect(update_user)

        app().init_translations({
            menu_bar.setWindowTitle: 'gui.menu_bar.title',
            status_bar.setWindowTitle: 'gui.status_bar.title',
            settings.setText: 'gui.menus.settings',
            tools.setText: 'gui.menus.tools',
            help.setText: 'gui.menus.help',
            account.setText: 'gui.menus.account',
            logger.label.setText: 'gui.status.default'
        })

        self.exception_reporter = logger.reporter

        self.addToolBar(Qt.ToolBarArea.TopToolBarArea, menu_bar)
        self.addToolBar(Qt.ToolBarArea.BottomToolBarArea, status_bar)

        EventBus['exceptions'].subscribe(lambda e: logger.label.setText(f'{e.exception}...'), ExceptionEvent)
        for action in (account, settings, tools, help):
            menu_bar.addSeparator()
            menu_bar.addAction(action)

        status_bar.addWidget(logger)
        status_bar.addSeparator()
        status_bar.addWidget(logger.label)

    # noinspection PyTypeChecker
    def _init_ui(self) -> None:
        """Initialize the UI, including Layouts and widgets."""
        def add_conversation():
            self.conversation_counter += 1
            self.conversation_tabs.addTab(ConversationView(), f'Conversation {self.conversation_counter}')

        def remove_conversation(index: int):
            view: ConversationView = self.conversation_tabs.widget(index)  # type: ignore
            view.deleteLater()

            self.conversation_tabs.removeTab(index)
            if not self.conversation_tabs.count():
                self.conversation_counter = 0
                add_conversation()

        # Define widget attributes
        # Cannot be defined in init_objects() as walrus operators are not allowed for object attribute assignment.
        # This works in the standard AST, but is a seemingly arbitrary limitation set by the interpreter.
        # See:
        # https://stackoverflow.com/questions/64055314/why-cant-pythons-walrus-operator-be-used-to-set-instance-attributes#answer-66617839
        self.conversation_tabs = QTabWidget(self)

        init_objects({
            self.conversation_tabs: {
                'tabsClosable': True,
                'tabCloseRequested': remove_conversation
            },

            (add_conversation_button := QPushButton(self.conversation_tabs)): {
                'size': {'fixed': (None, 26)},
                'icon': app().icon_store['add'],
                'clicked': add_conversation
            }

        })

        add_conversation()
        self.conversation_tabs.setCornerWidget(add_conversation_button, Qt.Corner.TopLeftCorner)

        init_layouts({
            # Main layout
            (layout := QVBoxLayout()): {
                'items': (self.conversation_tabs,)
            }
        })

        init_objects({
            (main_widget := QWidget()): {'layout': layout},

            self: {'centralWidget': main_widget}
        })

    # # # # # Events

    def show(self) -> None:
        """After window is displayed, show warnings if not already warned."""
        super().show()

        if app().first_launch:
            app().windows['readme_viewer'].show()
            app().show_dialog('information.first_launch', self)

        if not app().client.session_token:
            app().client.authenticationRequired.emit()

    def closeEvent(self, event: QCloseEvent) -> None:
        """Close all detached/children windows and quit application."""
        super().closeEvent(event)
        # Remember window size
        app().settings.reload()
        app().settings['gui/window/x_size'] = self.size().width()
        app().settings['gui/window/y_size'] = self.size().height()
        app().settings.save()

        app().quit()
        event.accept()



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Defines the ChatGPT-GUI HTTP Client."""
from __future__ import annotations

__all__ = (
    'Client',
)

import datetime as dt
import json
import os
from json import JSONDecodeError
from typing import Any
from uuid import UUID

from PySide6.QtCore import *

from ...constants import *
from ...models import CaseInsensitiveDict
from ...utils import decode_url
from ...utils import hide_windows_file
from ..manager import NetworkSession
from ..manager import Request
from ..manager import Response
from .auth import Authenticator
from .structures import Action
from .structures import Conversation
from .structures import Message
from .structures import User


_DATE_FORMAT: str = '%Y-%m-%dT%H:%M:%S.%fZ'


class Client(QObject):
    """Asynchronous HTTP REST Client that interfaces with ChatGPT."""

    authenticationRequired = Signal()
    receivedMessage = Signal(Message, Conversation)
    signedOut = Signal()

    def __init__(self, parent: QObject, **kwargs) -> None:
        """Initialize OpenAI Client.

        Order of precedence for session token::

            1. session_token kwarg value
            2. CHATGPT_SESSION_AUTH environment variable
            3. user's .session config file (Created by application)

        :keyword session_token: Session token, allows for creation of access tokens.
        """
        super().__init__(parent)
        self.receivedMessage.connect(lambda msg, convo: print(f'Conversation: {convo.uuid} | {msg}'))

        self.authenticator = Authenticator(self)
        self.authenticator.authenticationSuccessful.connect(self.new_session)

        self.conversations: dict[UUID, Conversation] = {}
        self.host: str = 'chat.openai.com'
        self.models: list[str] | None = None
        self.user: User | None = None

        self._session_expire: dt.datetime | None = None
        self._first_request: bool = True
        self._access_token: str | None = None

        # Load session token
        self._session_token: str | None = kwargs.pop('session_token', os.getenv('CHATGPT_SESSION_AUTH', None))
        if self._session_token is None and (CG_SESSION_PATH.is_file() or CG_SESSION_PATH_OLD.is_file()):
            # Read from file if value was not provided by kwarg or ENV
            # Migrate old format to new format.
            if CG_SESSION_PATH_OLD.is_file():
                hide_windows_file(CG_SESSION_PATH_OLD, unhide=True)

                # Read old token and dump in new format
                _token_old = CG_SESSION_PATH_OLD.read_text(encoding='utf8')
                CG_SESSION_PATH.write_text(json.dumps({
                    'user': {},
                    'expires': None,
                    'token': _token_old
                }, indent=2), encoding='utf8')

                # Delete old file and mark new file as hidden
                CG_SESSION_PATH_OLD.unlink()
                hide_windows_file(CG_SESSION_PATH)

            try:
                _session_data: dict[str, Any] = json.loads(CG_SESSION_PATH.read_text(encoding='utf8'))

            except JSONDecodeError:
                # Delete session file if invalid json.
                CG_SESSION_PATH.unlink()

            else:
                if user := _session_data.get('user'):
                    self.user = User.from_json(user)

                if expires := _session_data.get('expires'):
                    self._session_expire = dt.datetime.strptime(expires, _DATE_FORMAT)

                self._session_token = _session_data['token']

        self.session: NetworkSession = NetworkSession(self)
        self.session.headers = CaseInsensitiveDict({
            'Accept': '*/*',
            'Accept-Encoding': ', '.join(('gzip', 'deflate', 'br')),
            'Accept-Language': 'en-US,en;q=0.5',
            'Connection': 'keep-alive',
            'DNT': '1',
            'Host': self.host,
            'Origin': 'https://chat.openai.com',
            'Referer': 'https://chat.openai.com/chat',
            'Sec-Fetch-Dest': 'empty',
            'Sec-Fetch-Mode': 'cors',
            'Sec-Fetch-Site': 'same-site',
            'Sec-GPC': '1',
            'TE': 'trailers',
            'User-Agent': CG_USER_AGENT,
            'X-OpenAI-Assistant-App-Id': '',
        })

        if self.session_token:
            self.set_cookie('__Secure-next-auth.session-token', self.session_token)

    @property
    def api_root(self) -> str:
        """Root of sent API requests."""
        return f'https://{self.host}/'

    @property
    def access_token(self) -> str | None:
        """Bearer token to authenticate self to API endpoints."""
        return self._access_token

    @access_token.setter
    def access_token(self, value: str) -> None:
        value = decode_url(value)

        self._access_token = value.strip()
        self.session.headers['Authorization'] = f'Bearer {self._access_token}'

    @access_token.deleter
    def access_token(self) -> None:
        self._access_token = None
        if 'Authorization' in self.session.headers:
            self.session.headers.pop('Authorization')

    @property
    def session_token(self) -> str | None:
        """Bearer token to authenticate self to API endpoints."""
        return self._session_token

    @session_token.setter
    def session_token(self, value: str) -> None:
        value = decode_url(value)

        self._session_token = value.strip()
        self.set_cookie('__Secure-next-auth.session-token', value)

        hide_windows_file(CG_SESSION_PATH, unhide=True)
        CG_SESSION_PATH.write_text(json.dumps({
            'user': self.user.to_json() if self.user is not None else {},
            'expires': self._session_expire.strftime(_DATE_FORMAT) if self._session_expire is not None else None,
            'token': self._session_token
        }, indent=2), encoding='utf8')
        hide_windows_file(CG_SESSION_PATH)

    @session_token.deleter
    def session_token(self) -> None:
        self._session_token = None
        self._session_expire = None
        self.user = None
        self.delete_cookie('__Secure-next-auth.session-token')
        if CG_SESSION_PATH.is_file():
            CG_SESSION_PATH.unlink()

        self.signedOut.emit()

    def _get(self, path: str, update_auth_on_401: bool = True, **kwargs) -> Response:
        """Get a :py:class:`Response` from ChatGPT.

        :param path: path to append to the API root
        :param update_auth_on_401: run self._refresh_auth if response status code is 401 Unauthorized
        :param kwargs: Key word arguments to pass to the requests GET Request.
        """
        if self._first_request:
            self._first_request = False
            self._get('chat')
            self.refresh_auth()

        response: Response = self.session.get(self.api_root + path.strip(), wait_until_finished=True, **kwargs)

        if session_token := self.session.cookies.get('__Secure-next-auth.session-token'):
            self.session_token = session_token

        if response.code and not response.ok:
            # Handle errors
            if response.code == 401 and update_auth_on_401 and self.access_token is not None:
                if self.refresh_auth():
                    response = self._get(path, False, **kwargs)

        return response

    def get_models(self) -> list[dict[str, Any]] | None:
        """Get the list of models to use with ChatGPT.

        :return: List of model data. The "slug" key is the model name.
        """
        return self._get('backend-api/models').json.get('models')

    def send_message(self, message_text: str, conversation: Conversation) -> None:
        """Send a message and emit the AI's response through `the `receivedMessage`` signal.

        Automatically handles the last message id.

        :param message_text: Message to send.
        :param conversation: Conversation to send message in.
        :raises ValueError: If Response couldn't be parsed as a text/event-stream.
        """
        if not self.models:
            if not (models := self.get_models()):
                raise ValueError('Couldn\'t get model data from ChatGPT. Check to make sure you\'re authenticated.')
            self.models = [model['slug'] for model in models]

        if conversation.messages:
            parent_message: Message = conversation.messages[-1]
        else:
            parent_message = Message()

        message: Message = Message(text=message_text)
        action: Action = Action(self.models[0], conversation, [message], parent_message)

        request: Request = Request(
            'POST', self.api_root + 'backend-api/conversation',
            headers={'Accept': 'text/event-stream', 'Content-type': 'application/json'},
            json=action.to_json(), timeout=180.0,
        )

        response: Response = request.send(self.session, wait_until_finished=True)

        # Get the finished stream from the text/event-stream
        # Index -1 Is empty string, as the response ends with 2 newlines
        # Index -2 Is the "[DONE]", signifier
        # Index -3 Is the finished product
        try:
            last_stream: str = response.text.split('\n\n')[-3]
        except IndexError as e:
            raise ValueError(f'Message response is not a text/event-stream ({response.text}).') from e

        json_response: dict[str, Any] = json.loads(last_stream.removeprefix('data: ').strip())
        response_message: Message = Message.from_json(json_response['message'])

        conversation.messages.extend(action.messages)
        conversation.messages.append(response_message)

        if conversation.uuid is None:
            conversation.uuid = UUID(json_response['conversation_id'])

        if conversation.uuid not in self.conversations:
            self.conversations[conversation.uuid] = conversation

        self.receivedMessage.emit(response_message, conversation)

    def hidden_token(self) -> str:
        """:return: The first and last 3 characters of the session token, seperated by periods."""
        key = self.session_token
        if key is not None and len(key) > 6:
            return f'{key[:3]}{"." * 50}{key[-3:]}'
        return 'None'

    def refresh_auth(self) -> bool:
        """Refresh authentication to OpenAI servers.

        If session is invalid, ask for new credentials

        :return: True if successful, else False.
        """
        if not self.session_token or (self._session_expire is not None and self._session_expire < dt.datetime.now()):
            self.authenticationRequired.emit()

            # Delete expired session
            if self.session_token:
                del self.session_token

            return False

        response = self._get('api/auth/session')

        # Ignore next refresh if it has the same value
        if etag := response.headers.get('ETag'):
            self.session.headers['If-None-Match'] = etag

        if user := response.json.get('user'):
            self.user = User.from_json(user)

        if access_token := response.json.get('accessToken'):
            self.access_token = access_token

        if session_expire := response.json.get('expires'):
            self._session_expire = dt.datetime.strptime(session_expire, _DATE_FORMAT)

        if session_token := self.session.cookies.get('__Secure-next-auth.session-token'):
            self.session_token = session_token

        return True

    def signin(self, username: str, password: str) -> None:
        """Signin to OpenAI using the specified username and password.

        :param username: Username to signin to (email address).
        :param password: Password associated with username.
        """
        self.authenticator.username = username
        self.authenticator.password = password
        self.authenticator.authenticate()

    def new_session(self, session_token: str, *_) -> None:
        """Call with new session token provided by authenticator.

        :param session_token: New session token to use.
        """
        self.session_token = session_token
        self.refresh_auth()

    def delete_cookie(self, name: str) -> None:
        """Delete given cookie if cookie exists."""
        self.session.clear_cookies(self.host, '/', name)

    def set_cookie(self, name: str, value: str) -> None:
        """Set cookie value in Cookie jar."""
        self.session.set_cookie(name, value, self.host)



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""ExternalTextBrowser implementation."""
from __future__ import annotations

__all__ = (
    'ConversationView',
)

from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ...network.client import Conversation
from ...network.client import Message
from ...utils import init_layouts
from ...utils import init_objects
from ..aliases import app
from ..aliases import tr
from .external_text_browser import ExternalTextBrowser
from .paste_line_edit import PasteLineEdit


class ConversationView(QFrame):
    """Viewer for a ChatGPT conversation."""

    def __init__(self, conversation: Conversation | None = None, *args, **kwargs) -> None:
        """Initialize :py:class:`ConversationView` values."""
        super().__init__(*args, **kwargs)
        self.conversation = conversation if conversation is not None else Conversation()

        self._init_ui()

    def _init_ui(self) -> None:

        def toggle_multiline():
            nonlocal multiline_mode

            if multiline_mode := not multiline_mode:
                input_buttons.insertWidget(2, self.send_button)
                self.input_multi.setText(self.input_line.text())
                self.input_line.clear()
                self.input_line.setHidden(True)
                self.input_multi.setHidden(False)
            else:
                input_layout.insertWidget(2, self.send_button)
                self.input_line.setText(self.input_multi.toPlainText())
                self.input_multi.clear()
                self.input_multi.setHidden(True)
                self.input_line.setHidden(False)

        multiline_mode: bool = True
        self.output = ExternalTextBrowser(self)
        self.input_line = PasteLineEdit(self)
        self.input_multi = QTextEdit(self)
        self.send_button = QPushButton(self)
        self.toggle_multiline_button = QPushButton(self)
        font = QFont('Söhne', 11)

        init_objects({
            self.output: {
                'font': font,
                'placeholderText': tr('gui.output_text.placeholder'),
            },

            self.input_line: {
                'font': font,
                'placeholderText': tr('gui.input_field.placeholder'),
                'returnPressed': self.send_message
            },

            self.input_multi: {
                'font': font,
                'placeholderText': tr('gui.input_field.placeholder'),
            },

            self.send_button: {
                'size': {'fixed': (120, None)},
                'text': tr('gui.send_input', key_eval=False),
                'clicked': self.send_message
            },

            self.toggle_multiline_button: {
                'size': {'fixed': (120, None)},
                'icon': self.style().standardIcon(QStyle.StandardPixmap.SP_FileDialogDetailedView),
                'text': tr('gui.toggle_multiline', key_eval=False),
                'clicked': toggle_multiline
            }
        })

        init_layouts({
            (input_buttons := QHBoxLayout()): {
                'items': (
                    QSpacerItem(0, 0, hData=QSizePolicy.Policy.MinimumExpanding),
                    self.toggle_multiline_button,
                    self.send_button
                )
            },

            (input_layout := QHBoxLayout()): {
                'items': (self.input_line, self.input_multi)
            },

            QVBoxLayout(self): {
                'items': (self.output, input_layout, input_buttons)
            }

        })

        toggle_multiline()
        self.output.setPlaceholderText(tr('gui.output_text.placeholder'))

        app().client.receivedMessage.connect(self.receive_message)

    def append_to_view(self, text: str) -> None:
        """Append some new text to the output.

        Adds two newlines after the text for better differentiation between messages.
        Also scrolls down to bottom for you.

        :param text: Text to append to output.
        """
        self.output.setText(f'{self.output.toPlainText()}{text}\n\n')
        self.output.verticalScrollBar().setValue(self.output.verticalScrollBar().maximum())

    def send_message(self) -> None:
        """Send a message to the client using the current input text.

        Clears the input and disables after sending.
        """
        message: str = self.input_line.text() or self.input_multi.toPlainText()
        self.input_line.clear()
        self.input_multi.clear()
        self.send_button.setDisabled(True)
        self.append_to_view(tr('gui.output_text.you_prompt', message, key_eval=False))  # type: ignore

        app().client.send_message(message, self.conversation)  # type: ignore

    def receive_message(self, message: Message, conversation: Conversation) -> None:
        """Receive a response from the client.

        :param message: Message received.
        :param conversation: Conversation received in.
        """
        if conversation is self.conversation:
            self.send_button.setDisabled(False)
            self.append_to_view(tr('gui.output_text.ai_prompt', message.text, key_eval=False))



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""Account menu implementation."""
from __future__ import annotations

__all__ = (
    'AccountContextMenu',
    'confirm_sign_out',
)

from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ...utils import add_menu_items
from ...utils import init_objects
from ..aliases import app
from ..aliases import tr


def confirm_sign_out() -> None:
    """Confirm user intention and sign out.

    :raises ValueError: If user is None.
    """
    if (user := app().client.user) is None:
        raise ValueError('Cannot sign out from a null User.')

    confirmation: bool = app().show_dialog(
        'questions.sign_out', None,
        buttons=QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.Cancel,
        default_button=QMessageBox.StandardButton.Cancel,
        description_args=(user.email,)
    ).role == QMessageBox.ButtonRole.YesRole

    if confirmation:
        del app().client.session_token


class AccountContextMenu(QMenu):
    """Context menu that manages the logged in account."""

    def __init__(self, parent) -> None:
        """Create a new :py:class:`AccountContextMenu`."""
        super().__init__(parent)

        init_objects({
            (sign_in := QAction(self)): {
                'disabled': app().client.user is not None,
                'text': tr('gui.menus.account.sign_in'),
                'triggered': app().windows['sign_in'].show
            },

            (sign_out := QAction(self)): {
                'disabled': app().client.user is None,
                'text': tr('gui.menus.account.sign_out'),
                'triggered': confirm_sign_out
            },
        })

        add_menu_items(self, [
            'Sign In', sign_in,
            'Sign Out', sign_out
        ])



###################################################################################################
#                              MIT Licence (C) 2022 Cubicpath@Github                              #
###################################################################################################
"""PasteLineEdit implementation."""
from __future__ import annotations

__all__ = (
    'SignInDialog',
)

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *

from ...models import DeferredCallable
from ...utils import init_layouts
from ...utils import init_objects
from ..aliases import app


class SignInDialog(QWidget):
    """A :py:class:`SignInDialog` which signs in to the ChatGPT Client."""

    def __init__(self) -> None:
        """Create a new :py:class:`SignInDialog`."""
        super().__init__(None)

        app().client.authenticator.authenticationSuccessful.connect(self.close)
        app().client.authenticator.authenticationFailed.connect(DeferredCallable(self._reset_fields))

        self._init_ui()

    def _init_ui(self) -> None:
        username_label = QLabel()
        password_label = QLabel()
        self.username_input = QLineEdit(self)
        self.password_input = QLineEdit(self)
        self.set_username_button = QPushButton(self)
        self.set_password_button = QPushButton(self)
        self.sign_in_button = QPushButton(self)

        init_objects({
            self: {
                'size': {'fixed': (500, 200)},
                'windowFlags': Qt.WindowType.Dialog,
                'windowModality': Qt.WindowModality.ApplicationModal,
                'windowIcon': app().icon_store['account'],
            },

            self.username_input: {
                'returnPressed': self._enter_username,
            },

            self.password_input: {
                'echoMode': QLineEdit.EchoMode.Password,
                'returnPressed': self._enter_password,
            },

            self.sign_in_button: {
                'clicked': DeferredCallable(
                    app().client.signin,
                    self.username_input.text,
                    self.password_input.text
                ),
            },

            self.set_username_button: {
                'size': {'minimum': (45, None)},
                'clicked': self._enter_username,
            },

            self.set_password_button: {
                'size': {'minimum': (45, None)},
                'clicked': self._enter_password,
            },

            (reset_fields_button := QPushButton(self)): {
                'clicked': self._reset_fields,
            },

            (cancel_button := QPushButton(self)): {
                'clicked': self.close,
            }
        })

        app().init_translations({
            self.setWindowTitle: 'gui.sign_in_dialog.title',
            username_label.setText: 'gui.sign_in_dialog.username_label',
            password_label.setText: 'gui.sign_in_dialog.password_label',
            self.username_input.setPlaceholderText: 'gui.sign_in_dialog.username_placeholder',
            self.password_input.setPlaceholderText: 'gui.sign_in_dialog.password_placeholder',

            self.set_username_button.setText: 'gui.sign_in_dialog.set_field',
            self.set_password_button.setText: 'gui.sign_in_dialog.set_field',
            self.sign_in_button.setText: 'gui.sign_in_dialog.sign_in',
            reset_fields_button.setText: 'gui.sign_in_dialog.reset_fields',
            cancel_button.setText: 'gui.sign_in_dialog.cancel'
        })

        init_layouts({
            (input_layout := QGridLayout()): {
                'items': (
                    (username_label, 1, 0),
                    (self.username_input, 1, 1),
                    (self.set_username_button, 1, 2),
                    (password_label, 2, 0),
                    (self.password_input, 2, 1),
                    (self.set_password_button, 2, 2)
                )
            },

            (bottom := QHBoxLayout()): {
                'items': (reset_fields_button, cancel_button)
            },

            # Main layout
            QVBoxLayout(self): {
                'items': (input_layout, self.sign_in_button, bottom)
            }
        })

        input_layout.setRowMinimumHeight(2, 40)
        input_layout.setRowStretch(3, 1)
        input_layout.setColumnMinimumWidth(1, 200)
        input_layout.setSpacing(5)

    def _enter_username(self) -> None:
        self.username_input.setDisabled(True)
        self.set_username_button.setDisabled(True)

        self.password_input.setDisabled(False)
        self.set_password_button.setDisabled(False)
        self.password_input.setFocus()

    def _enter_password(self) -> None:
        self.password_input.setDisabled(True)
        self.set_password_button.setDisabled(True)

        self.sign_in_button.setDisabled(False)
        self.sign_in_button.setFocus()

    def _reset_fields(self) -> None:
        self.username_input.clear()
        self.password_input.clear()
        self.username_input.setDisabled(False)
        self.password_input.setDisabled(True)

        self.set_username_button.setDisabled(False)
        self.set_password_button.setDisabled(True)

        self.sign_in_button.setDisabled(True)

    def showEvent(self, event: QShowEvent) -> None:
        """Reset fields when dialog is shown to user."""
        super().showEvent(event)
        self._reset_fields()
